#include <iostream>
#include <sstream>
#include <iomanip>
#include <vector>
#include <string>
#include <unistd.h>
#include "pvo.hh"
#include "pvo-metadata.hh"
#include "pvo-FITS.hh"
#include "pvo-utils.hh"

namespace PVO
{

  namespace Meta
  {

    Metadata::Metadata(){
      sepFlag=false; 
      angsep=nullFloat;
    }

    Metadata::~Metadata()
    {
    }

    Metadata::Metadata(const Metadata& m)
    {
      filename        = m.filename      ;
      fileloc	      = m.fileloc	      ;
      headerversion   = m.headerversion ;
      file_size_bytes = m.file_size_bytes;
      survey	      = m.survey	      ;
      srcname	      = m.srcname	      ;
      telescope	      = m.telescope     ;	    
      data_type	      = m.data_type     ;	    
      raj	      = m.raj	      ;
      decj	      = m.decj	      ;
      rajd	      = m.rajd	      ;
      decjd	      = m.decjd	      ;
      gl	      = m.gl	      ;
      gb	      = m.gb	      ;
      date	      = m.date	      ;
      beamsize	      = m.beamsize      ;	    
      bmaj	      = m.bmaj	      ;
      bmin	      = m.bmin	      ;
      bpa	      = m.bpa	      ;
      mjd	      = m.mjd	      ;
      obsfreq	      = m.obsfreq	      ;
      tsamp	      = m.tsamp	      ;
      scanlen	      = m.scanlen	      ;
      nbits	      = m.nbits	      ;
      nchan	      = m.nchan	      ;
      npol 	      = m.npol	      ;
      nbin 	      = m.nbin	      ;
      nsub	      = m.nsub	      ;
      chanbw	      = m.chanbw	      ;
      angsep	      = m.angsep	      ;
      sepFlag         = m.sepFlag       ;

    }

    Metadata& Metadata::operator= (const Metadata& m)
    {
      if(this == &m) return *this;   // Gracefully handle self assignment
      this->filename      = m.filename      ;
      this->fileloc	    = m.fileloc	      ;
      this->file_size_bytes = m.file_size_bytes;
      this->headerversion = m.headerversion ;
      this->survey	    = m.survey	      ;
      this->srcname	    = m.srcname	      ;
      this->telescope	    = m.telescope     ;	    
      this->data_type	    = m.data_type     ;	    
      this->raj	    = m.raj	      ;
      this->decj	    = m.decj	      ;
      this->rajd	    = m.rajd	      ;
      this->decjd	    = m.decjd	      ;
      this->gl	    = m.gl	      ;
      this->gb	    = m.gb	      ;
      this->date	    = m.date	      ;
      this->beamsize	    = m.beamsize      ;	    
      this->bmaj	    = m.bmaj	      ;
      this->bmin	    = m.bmin	      ;
      this->bpa	    = m.bpa	      ;
      this->mjd	    = m.mjd	      ;
      this->obsfreq	    = m.obsfreq	      ;
      this->tsamp	    = m.tsamp	      ;
      this->scanlen	    = m.scanlen	      ;
      this->nbits	    = m.nbits	      ;
      this->nchan	    = m.nchan	      ;
      this->npol 	    = m.npol	      ;
      this->nbin 	    = m.nbin	      ;
      this->nsub	    = m.nsub	      ;
      this->chanbw	    = m.chanbw	      ;
      this->angsep	    = m.angsep	      ;
      this->sepFlag        = m.sepFlag       ;

      return *this;
    }
    
    

    void Metadata::define(FITS::PSRFits fitsfile)
    {
      /**
       * Define a set of metadata.
       *
       * This takes a FITS file and reads in all desired metadata
       * fields from it. If certain fields are not present, the
       * appropriate NULL value is recorded (such as Meta::nullString
       * or Meta::nullFloat).
       *
       * \param fitsfile The PSRFits object containing the FITS file
       *                  information.
       */

      this->filename = fitsfile.getFITSname();

      int pos = this->filename.find("_");
      std::string extension="";
      std::string surveyStr = this->filename.substr(0,pos);
      size_t dot = this->filename.rfind('.');
      if(dot!=std::string::npos) 
	extension = this->filename.substr(dot,this->filename.size());
      this->survey = nullString;
      if(extension == ".sf"){
	if(surveyStr == "S70") this->survey = "PKS70";
	else if(surveyStr == "PM") this->survey = "PM";
	else if(surveyStr == "PH") this->survey = "PH";
      }

      int cwdSize=200;
      char *cwd = new char[200];
      if(this->filename[0]!='/'){ 
	// if the directory doesn't start with a /, it is the current
	// directory.
	getcwd(cwd,cwdSize);
	this->fileloc = cwd;
	this->fileloc += "/";
      }
      this->fileloc += this->filename;

      std::stringstream sysmsg;
      sysmsg << "du -bL " << this->fileloc << " | awk '{print $1}' > /tmp/size";
      system(sysmsg.str().c_str());
      std::ifstream fin("/tmp/size");
      fin >> this->file_size_bytes;
      fin.close();      

      fitsfile.gotoPHU();

      if(fitsfile.getKeyword<std::string>("HDRVER",this->headerversion)!=0) 
	this->headerversion = nullString;
      this->headerversion = deblank(this->headerversion);

      if(fitsfile.getKeyword<std::string>("SRC_NAME",this->srcname)!=0) 
	this->srcname = nullString;
      else this->srcname = deblank(this->srcname);
      if(this->srcname=="") this->srcname = nullString;

      if(fitsfile.getKeyword<std::string>("TELESCOP",this->telescope)!=0) 
	this->telescope = nullString;
      else this->telescope = deblank(this->telescope);
      if(this->telescope=="") this->telescope = nullString;

      if(fitsfile.getKeyword<std::string>("BACKEND",this->data_type)!=0) 
	this->data_type = nullString;
      else this->data_type = deblank(this->data_type);
      if(this->data_type=="") this->data_type = nullString;

      if(fitsfile.getKeyword<std::string>("RA",this->raj)!=0) 
	this->raj = nullString;
      else this->rajd = dmsToDec(this->raj) * 15.;

      if(fitsfile.getKeyword<std::string>("DEC",this->decj)!=0) 
	this->decj = nullString;
      else this->decjd = dmsToDec(this->decj);

      if(this->raj != nullString && this->decj != nullString)
	equatToGal(this->rajd,this->decjd, this->gl, this->gb);
      else
	this->gl = this->gb = nullFloat;

      if(fitsfile.getKeyword<std::string>("DATE-OBS",this->date)!=0) 
	this->date = nullString;

      if(fitsfile.getKeyword<float>("OBSFREQ",this->obsfreq)!=0) 
	this->obsfreq = nullFloat;
  
      if(fitsfile.getKeyword<float>("BMAJ",this->bmaj)==0){
	fitsfile.getKeyword<float>("BMIN",this->bmin);
	fitsfile.getKeyword<float>("BPA",this->bpa);
	this->beamsize = 0.5*(bmaj+bmin); 
      }
      else this->bmaj = this->bmin = this->bpa = this->beamsize = nullFloat;

      if(fitsfile.getKeyword<float>("SCANLEN",this->scanlen)!=0) 
	this->scanlen = nullFloat;

      float imjd,smjd;
      if(fitsfile.getKeyword<float>("STT_IMJD",imjd)!=0) 
	this->mjd = nullFloat;
      else{
	fitsfile.getKeyword<float>("STT_SMJD",smjd);
	this->mjd = imjd + smjd / (24.*3600.);
      }

      ////////////////////
      // From here, we get information from the HISTORY table.
      fitsfile.gotoHistory();

      if(fitsfile.readColumnValue("NCHAN",1,&this->nchan)!=0) 
	this->nchan = nullInt;
      if(this->nchan == 0) this->nchan = nullInt;

      if(fitsfile.readColumnValue("CHAN_BW",1,&this->chanbw)!=0) 
	this->chanbw = nullFloat;
      if(!(fabs(this->chanbw)>0.)) this->chanbw = nullFloat; // only set to null if it is zero.

      ////////////////////
      // From here, we get information from the SUBINT table.
      fitsfile.gotoSubint();
      // TSAMP in seconds.
      //      if(fitsfile.readColumnValue("TBIN",1,&this->tbin)!=0) 
      if(fitsfile.getKeyword<float>("TSAMP",this->tsamp)!=0) 
	this->tsamp = nullFloat;
      if(!(this->tsamp>0.)) this->tsamp = nullFloat;

      // number of bins
      if(fitsfile.getKeyword<int>("NBIN",this->nbin)!=0) 
	this->nbin = nullInt;
      if(!(this->nbin>0.)) this->nbin = nullInt;

      // number of polarisations
      if(fitsfile.getKeyword<int>("NPOL",this->npol)!=0) 
	this->npol = nullInt;
      if(!(this->npol>0.)) this->npol = nullInt;

      // number of bits
      if(fitsfile.getKeyword<int>("NBITS",this->nbits)!=0) 
	this->nbits = nullInt;
      if(!(this->nbits>0.)) this->nbits = nullInt;

      // number of subintegrations
      if(fitsfile.getKeyword<int>("NAXIS2",this->nsub)!=0) 
	this->nsub = nullInt;
      if(!(this->nsub>0.)) this->nsub = nullInt;


    }
    //----------------------------------------------------
    void Metadata::setAngSep(float ra, float dec)
    {
      /**
       * Set the value of the Metadata::angsep parameter, being the
       * angular separation from a given (ra, dec) position, in
       * degrees.
       *
       * \param ra The right ascension of the base position.
       * \param dec The declination of the base position.
       */ 
      this->angsep = angSep(ra, dec, rajd, decjd); 
      this->sepFlag=true;
    }
    //----------------------------------------------------

    bool Metadata::close(float radius)
    {
      /**
       * Uses the metadata quantity Meta::Metadata::angsep to
       * determine if the current object is within the given radius of
       * some position. The angsep quantity needs to have been defined
       * previously.
       *
       * \param radius The maximum separation, in degrees, that will
       *                return true.
       * \return True if angsep is less than the radius plus half the
       *          beamsize. False otherwise.
       */
      if(this->angsep < 0.) return false;
      else return (this->angsep < (radius + 0.5*this->beamsize));
    }
    //----------------------------------------------------

    bool Metadata::close(float ra, float dec, float radius)
    {
      /**
       * Determines whether the current object is within the given
       * radius of a given position. This function defines the
       * Metadata::angsep quantity in the process. 
       *
       * \param ra The right ascension of the base position.
       * \param dec The declination of the base position.
       * \param radius The maximum separation, in degrees, from the
       *                base position that will return true.
       * \return True if angsep is less than the radius plus half the
       *          beamsize. False otherwise.
       */
      this->setAngSep(ra,dec);
      return (angsep < (radius+0.5*this->beamsize));
    }
    //----------------------------------------------------

    std::string floatToString(float f)
    {
      /**
       *  Convert a float value to a string, using the default
       *  precision of Meta::fltwidth. If the value is null, return
       *  the null string.
       *  \param f The float value.
       *  \return A string showing the value, or the null string if
       *           appropriate.
       */
      if(f > nullFloat){
	std::stringstream ss;
	ss << std::fixed << std::setprecision(fltwidth) << f;
	return ss.str();
      }
      else return nullString;
    }
    //----------------------------------------------------

    std::string doubleToString(double d)
    {
      /**
       *  Convert a double value to a string, using the default
       *  precision of Meta::dubwidth. If the value is null, return
       *  the null string.
       *  \param d The double value.
       *  \return A string showing the value, or the null string if
       *           appropriate.
       */
      if(d > nullFloat){
	std::stringstream ss;
	ss << std::fixed << std::setprecision(dubwidth) << d;
	return ss.str();
      }
      else return nullString;
    }
    //----------------------------------------------------

    std::string doubleToString(double d, int prec)
    {
      /**
       *  Convert a double value to a string, using the precision of
       *  prec. If the value is null, return the null string.
       *  \param d The double value.
       *  \param prec The desired precision.
       *  \return A string showing the value, or the null string if
       *           appropriate.
       */
      if(d > nullFloat){
	std::stringstream ss;
	ss << std::fixed << std::setprecision(prec) << d;
	return ss.str();
      }
      else return nullString;
    }
    //----------------------------------------------------

    std::string intToString(int i)
    {
      /**
       *  Convert an int value to a string. If the value is null,
       *  return the null string.
       *  \param i The int value.
       *  \return A string showing the value, or the null string if
       *           appropriate.
       */
      if(i > nullInt){
	std::stringstream ss;
	ss << i;
	return ss.str();
      }
      else return nullString;
    }
    //----------------------------------------------------

    void printMetadataShort(std::ostream &theStream, Metadata data)
    {
      /**
       * A front-end to printMetadataShort(std::ostream,
       * std::vector<Metadata>).
       */
      std::vector<Metadata> datalist;
      datalist.push_back(data);
      printMetadataShort(theStream, datalist);
    }
    //----------------------------------------------------

    void printMetadataShort(std::ostream &theStream, std::vector<Metadata> datalist)
    {
      /**
       * Prints out the summary list of metadata to the desired
       * ostream. The summary list is a short list of columns,
       * designed for easy reading on a terminal. The columns are:
       * FILENAME, SRC_NAME, RAJ, DECJ, BEAM FWHM, OBSFREQ, SCANLEN,
       * DATE-OBS. The widths of the columns are fixed, except for the
       * first two, which expand in size to fit the data.
       *
       * \param theStream Where the data should be printed.
       * \param datalist The vector of Metadata objects that describe
       *                  the metadata.
       */
      int filenameSize = 8; // size of "FILENAME"
      int srcNameSize = 8;  // size of "SRC_NAME"
      int width;
      unsigned int i;
      bool showSep = false;
      if(datalist.size()>0){
	showSep = datalist[0].showSep();
      }
      for(i=0;i<datalist.size();i++){
	filenameSize = std::max(filenameSize,
				int(datalist[i].getFilename().size()));
	srcNameSize = std::max(srcNameSize,
			       int(datalist[i].getSrcname().size()));
      }
      width = filenameSize+srcNameSize+2+15+15+31+11+11+11;

      theStream << std::setfill('-')  
		<< std::setw(width) << "-";
      if(showSep) theStream << std::setw(13) << "-";
      
      theStream << "\n" << std::setfill(' ') 
		<< std::setw(filenameSize) << "FILENAME"  << " "
		<< std::setw(srcNameSize)  << "SRC_NAME"  << " "
		<< std::setw(14)           << "RAJ"       << " "
		<< std::setw(14)           << "DECJ"      << " "
		<< std::setw(10)           << "BEAM FWHM" << " "
		<< std::setw(10)           << "OBSFREQ"   << " "
		<< std::setw(10)           << "SCANLEN"   << " "
		<< std::setw(30)           << "DATE-OBS";
      if(showSep)
	theStream  << " " << std::setw(12) << "OFFSET"    << "\n";
      else theStream << "\n";

      theStream << std::setw(filenameSize) << ""          << " "
		<< std::setw(srcNameSize)  << ""          << " "
		<< std::setw(14)           << "[deg]"     << " "
		<< std::setw(14)           << "[deg]"     << " "
		<< std::setw(10)           << "[deg]"     << " "
		<< std::setw(10)           << "[MHz]"     << " "
		<< std::setw(10)           << "[s]"       << " "
		<< std::setw(30)           << "";
      if(showSep)
	theStream << " " << std::setw(12)  << "[deg]" << "\n";
      else  theStream << "\n";

      theStream << std::setfill('-')  
		<< std::setw(width) << "-";
      if(showSep) theStream << std::setw(13) << "-";
      theStream << "\n"<< std::setfill(' ') ;
    
      for(i=0;i<datalist.size();i++){

	theStream << std::setw(filenameSize) 
		  << datalist[i].getFilename() << " ";
	theStream << std::setw(srcNameSize) 
		  << datalist[i].getSrcname() << " ";

	if(datalist[0].showSep())
	  theStream << std::setprecision(6)
		    << std::setw(14)<< datalist[i].getRaD() << " "
		    << std::setw(14)<< datalist[i].getDecD()<< " ";
	else
	  theStream << std::setw(14)<< datalist[i].getRaS() << " "
		    << std::setw(14)<< datalist[i].getDecS()<< " ";

	theStream << std::setprecision(4)
		  << std::setw(10)<< floatToString(datalist[i].getBeamsize())<< " "
		  << std::setw(10)<< floatToString(datalist[i].getObsfreq()) << " "
		  << std::setw(10)<< floatToString(datalist[i].getScanlen()) << " "
		  << std::setw(30)<< datalist[i].getDate();

	if(datalist[0].showSep()) 
	  theStream << " " 
		    << std::setw(12)<< floatToString(datalist[i].getAngSep())<< "\n";
	else theStream << "\n";
	
      }


    }
    //----------------------------------------------------

    void printMetadataFull(std::string filename, Metadata data)
    {
      /**
       * A front-end to printMetadataFull(std::string,
       * std::vector<Metadata>).
       */
      std::vector<Metadata> datalist;
      datalist.push_back(data);
      printMetadataFull(filename, datalist);
    }
    //----------------------------------------------------

    std::string enclose(std::string input)
    {
      std::string output;
      if(input == nullString) output = nullString;
      else output = "\"" + input + "\"";
      return output;
    }

    void printMetadataFull(std::string filename, std::vector<Metadata> datalist)
    {
      /**
       * Prints out the full list of metadata for a set of objects to
       * the desired file given by filename. The metadata are printed
       * in csv (comma-separated values) format, and all metadata are
       * printed.
       *
       * \param filename The name of the file to which the data should
       *                  be printed.
       * \param datalist The vector of Metadata objects that describe
       *                  the metadata.
       */
      std::ofstream fileout(filename.c_str());

      unsigned int i;
    
      fileout << "Filename,"
	      << "\"File Location\","
	      << "File_size_bytes,"
	      << "\"Project ID\","
	      << "\"Header version\","
	      << "Survey,"
	      << "Telescope,"
	      << "Site,"
	      << "\"Source Name\","
	      << "\"UT Date\","
	      << "\"UT Time\","
	      << "MJD,"
	      << "RAJ,"
	      << "DECJ,"
	      << "RAJd,"
	      << "DECJd,"
	      << "GL,"
	      << "GB,"
	      << "BMAJ,"
	      << "BMIN,"
	      << "BPA,"
	      << "DM,"
	      << "Period,"
	      << "Config,"
	      << "Instrument,"
	      << "Data_type,"
	      << "Receiver,"
	      << "Observers,"
	      << "Obsfreq,"
	      << "ChanBW,"
	      << "NChan,"
	      << "NPol,"
	      << "NBin,"
	      << "NSub,"
	      << "\"Scan length\","
	      << "Tsamp,"
	      << "Nbits,"
	      << "Nbeam";
	    
      if(datalist[0].showSep())	fileout  << ",Offset\n";
      else fileout << "\n";
     
      for(i=0;i<datalist.size();i++){

	std::string date,time,obsdate;
	obsdate = datalist[i].getDate();
	size_t tloc = obsdate.find('T');
	date = obsdate.substr(0,tloc);
	time = obsdate.substr(tloc,obsdate.size()-tloc);

	fileout << enclose(datalist[i].getFilename() )      << ","
		<< enclose(datalist[i].getFileloc())        << ","
	        << datalist[i].getFileSize()                << ","
		<< nullString                               << "," //project_id
		<< enclose(datalist[i].getHeaderVersion())  << ","
		<< enclose(datalist[i].getSurvey())         << ","
		<< enclose(datalist[i].getTelescope())      << ","
		<< nullString                               << "," // site
		<< enclose(datalist[i].getSrcname())        << ","
		<< date                                     << ","
		<< time                                     << ","
		<< doubleToString(datalist[i].getMJD())     << ","
		<< enclose(datalist[i].getRaS())            << ","
	       	<< enclose(datalist[i].getDecS())           << ","
	        << std::fixed << std::setprecision(6)
	       	<< datalist[i].getRaD()                     << ","
	       	<< datalist[i].getDecD()                    << ","
	       	<< doubleToString(datalist[i].getGL(),6)    << ","
	       	<< doubleToString(datalist[i].getGB(),6)    << ","
		<< floatToString(datalist[i].getBmaj())     << ","
		<< floatToString(datalist[i].getBmin())     << ","
		<< floatToString(datalist[i].getBpa())      << ","
		<< nullString                               << "," // DM
		<< nullString                               << "," // Period
		<< nullString                               << "," // Config
		<< enclose(datalist[i].getDataType())       << "," // Instrument
		<< enclose(datalist[i].getDataType())       << ","
		<< nullString                               << "," // Receiver
		<< nullString                               << "," // Observers
		<< floatToString(datalist[i].getObsfreq())  << ","
		<< floatToString(datalist[i].getChanbw())   << ","
		<< intToString(datalist[i].getNchan())      << ","
		<< intToString(datalist[i].getNpol())       << ","
		<< intToString(datalist[i].getNbin())       << ","
		<< intToString(datalist[i].getNsub())       << ","
		<< floatToString(datalist[i].getScanlen())  << ","
		<< floatToString(datalist[i].getTsamp())    << ","
		<< intToString(datalist[i].getNbits())      << ","
		<< 1; //nbeam      

	if(datalist[0].showSep()) 
	  fileout << "," << floatToString(datalist[i].getAngSep()) << "\n";
	else fileout << "\n";

      }

      fileout.close();

    }

  }// end namespace Meta

} // end namespace PVO
